from pyadjoint import OverloadedType
from pyadjoint.tape import (
    annotate_tape,
    no_annotations,
    stop_annotating,
    get_working_tape,
)
from pyadjoint.overloaded_type import create_overloaded_object
from pyadjoint.enlisting import Enlist
from pyadjoint_utils import Block


def overload_point_map_call(func):
    assert (
        func.__name__ == "__call__"
    ), "this decorator only works on the __call__ function"

    def decorator(*args, **kwargs):
        annotate = annotate_tape(kwargs)
        if annotate:
            b_kwargs = PointMapBlock.pop_kwargs(kwargs)
            b_kwargs.update(kwargs)

        with stop_annotating():
            output = func(*args, **kwargs)

        output = Enlist(output)
        r = [create_overloaded_object(o) for o in output]

        if annotate:
            block = PointMapBlock(*args, output.delist(r))
            get_working_tape().add_block(block)
        return output.delist(r)

    return decorator


def overloaded_point_map(cls):
    cls.__call__ = overload_point_map_call(cls.__call__)
    return cls


class PointMapBlock(Block):
    def __init__(self, amap, arg, output, **kwargs):
        super(PointMapBlock, self).__init__()
        self._amap = amap
        args = Enlist(arg)
        outputs = Enlist(output)

        is_overloaded = [isinstance(output, OverloadedType) for output in outputs]
        num_overloaded = sum(is_overloaded)
        overloaded_map = [sum(is_overloaded[0:i]) for i in range(num_overloaded)]
        self._out_map = overloaded_map

        is_overloaded = [isinstance(arg, OverloadedType) for arg in args]
        num_overloaded = sum(is_overloaded)
        overloaded_map = [sum(is_overloaded[0:i]) for i in range(num_overloaded)]
        self._arg_map = overloaded_map

        self._ag_outputs = outputs
        for argi in self._arg_map:
            self.add_dependency(args[argi])

        self._args = args
        for outi in self._out_map:
            self.add_output(outputs[outi].block_variable)

    def __repr__(self):
        return f"PointMapBlock({self._amap!r})"

    @no_annotations
    def recompute(self, markings=False):
        outputs = self.get_outputs()

        inputs = [bv.saved_output for bv in self.get_dependencies()]
        relevant_outputs = [
            (i, bv) for i, bv in enumerate(outputs) if bv.marked_in_path or not markings
        ]

        if len(relevant_outputs) <= 0:
            return

        full_arg = sparse_insert(
            self._amap.source, self._args, inputs, self._arg_map, copy=True
        )
        new_outputs = Enlist(self._amap(full_arg))

        for idx, out in relevant_outputs:
            output = new_outputs[self._out_map[idx]]
            out.checkpoint = output

    @no_annotations
    def evaluate_adj(self, markings=False):
        outputs = self.get_outputs()
        adj_inputs = []
        has_input = False
        for output in outputs:
            adj_inputs.append(output.adj_value)
            if output.adj_value is not None:
                has_input = True

        if not has_input:
            return

        deps = self.get_dependencies()
        inputs = [bv.saved_output for bv in deps]
        relevant_dependencies = [
            (i, bv)
            for i, bv in enumerate(deps)
            if (bv.marked_in_path or not markings) and (bv not in outputs)
        ]
        # TODO: Why did I add (bv not in outputs) as a condition here? When would that ever happen, and why?

        if len(relevant_dependencies) <= 0:
            return

        full_arg = sparse_insert(
            self._amap.source, self._args, inputs, self._arg_map, copy=True
        )
        full_adj_vec = sparse_insert(
            self._amap.target, self._ag_outputs, adj_inputs, self._out_map
        )
        adj_outputs = Enlist(self._amap.adjoint(full_arg, full_adj_vec))

        for idx, dep in relevant_dependencies:
            adj_output = adj_outputs[self._arg_map[idx]]
            if adj_output is not None:
                dep.add_adj_output(adj_output)

    @no_annotations
    def evaluate_tlm(self, markings=False):
        deps = self.get_dependencies()
        tlm_inputs = []
        has_input = False
        for dep in deps:
            tlm_inputs.append(dep.tlm_value)
            if dep.tlm_value is not None:
                has_input = True

        if not has_input:
            return

        outputs = self.get_outputs()
        inputs = [bv.saved_output for bv in deps]
        relevant_outputs = [
            (i, bv)
            for i, bv in enumerate(outputs)
            if (bv.marked_in_path or not markings) and (bv not in deps)
        ]

        if len(relevant_outputs) <= 0:
            return

        full_arg = sparse_insert(
            self._amap.source, self._args, inputs, self._arg_map, copy=True
        )
        full_tlm_vec = sparse_insert(
            self._amap.source, self._args, tlm_inputs, self._arg_map
        )
        tlm_outputs = Enlist(self._amap.tlm(full_arg, full_tlm_vec))

        for idx, out in relevant_outputs:
            tlm_output = tlm_outputs[self._out_map[idx]]
            if tlm_output is not None:
                out.add_tlm_output(tlm_output)

    @no_annotations
    def evaluate_hessian(self, markings=False):
        outputs = self.get_outputs()
        hessian_inputs = []
        adj_inputs = []
        has_input = False
        for output in outputs:
            hessian_inputs.append(output.hessian_value)
            adj_inputs.append(output.adj_value)
            if output.hessian_value is not None:
                has_input = True

        if not has_input:
            return

        deps = self.get_dependencies()
        tlm_inputs = []
        for dep in deps:
            tlm_inputs.append(dep.tlm_value)

        inputs = [bv.saved_output for bv in deps]
        relevant_dependencies = [
            (i, bv) for i, bv in enumerate(deps) if bv.marked_in_path or not markings
        ]

        if len(relevant_dependencies) <= 0:
            return

        full_arg = sparse_insert(
            self._amap.source, self._args, inputs, self._arg_map, copy=True
        )
        full_adj_vec = sparse_insert(
            self._amap.target, self._ag_outputs, adj_inputs, self._out_map
        )
        full_hes_vec = sparse_insert(
            self._amap.target, self._ag_outputs, hessian_inputs, self._out_map
        )
        full_tlm_vec = sparse_insert(
            self._amap.source, self._args, tlm_inputs, self._arg_map
        )
        adj_outputs = Enlist(self._amap.adjoint(full_arg, full_hes_vec))
        hes_outputs = Enlist(self._amap.hessian(full_arg, full_adj_vec, full_tlm_vec))
        for i in range(len(adj_outputs)):
            hes_outputs[i] = hes_outputs[i] + adj_outputs[i]

        for idx, dep in relevant_dependencies:
            hessian_output = hes_outputs[self._arg_map[idx]]
            if hessian_output is not None:
                dep.add_hessian_output(hessian_output)

    @no_annotations
    def evaluate_tlm_matrix(self, markings=False):
        deps = self.get_dependencies()
        tlm_inputs = []
        has_input = False
        for dep in deps:
            tlm_inputs.append(dep.tlm_matrix)
            if dep.tlm_matrix is not None:
                has_input = True

        if not has_input:
            return

        outputs = self.get_outputs()
        inputs = [bv.saved_output for bv in deps]
        relevant_outputs = [
            (i, bv)
            for i, bv in enumerate(outputs)
            if (bv.marked_in_path or not markings) and (bv not in deps)
        ]

        if len(relevant_outputs) <= 0:
            return

        full_arg = sparse_insert(
            self._amap.source, self._args, inputs, self._arg_map, copy=True
        )
        full_tlm_vec = sparse_insert(
            self._amap.source, self._args, tlm_inputs, self._arg_map
        )
        full_tlm_vec = tlm_inputs
        tlm_outputs = Enlist(self._amap.tlm_matrix(full_arg, full_tlm_vec))
        # TODO: currently, it is assumed that this map has only one output.

        tlm_outputs = [tlm_outputs]
        for idx, out in relevant_outputs:
            idx_out = self._out_map[idx]
            tlm_matrix = tlm_outputs[idx_out]
            if tlm_matrix is not None:
                out.add_tlm_matrix(tlm_matrix)


def sparse_insert(space, near, sparse_vals, idxmap, copy=False):
    if not copy:
        full_vals = Enlist(space.point(zero=True, near=near.delist(tuple(near))))
    else:
        full_vals = near.copy()
    for i, val in enumerate(sparse_vals):
        if val is not None:
            full_vals[i] = val
    return near.delist(tuple(full_vals))



from crikit.covering import Covering, register_covering
from crikit.cr.numpy import Ndarrays
from crikit.cr.quadrature import get_quadrature_params, make_quadrature_spaces
from crikit.cr.types import PointMap
from crikit.cr.map_builders import CompositePointMap, IdentityPointMap, ParallelPointMap
from crikit.cr.ufl import UFLExprSpace, UFLFunctionSpace

from crikit.fe import *
from crikit.fe_adjoint import *
from crikit.projection import project
from pyadjoint_utils.fenics_adjoint import function_get_local, function_set_local


class Numpy_To_UFLFunctionSpace(PointMap):
    """This class is a point map that maps a NumPy array to a UFL function space.

    The constructor creates a quadrature space for the target space, using the
    given quadrature parameters. The ``__call__`` method sticks the given array
    into the quadrature space and projects it to the target function space.

    """

    def __init__(self, source, target, quad_space=None, quad_params=None):
        if quad_space is None:
            quad_space, quad_params = make_quadrature_spaces(
                target, quad_params=quad_params
            )
        self._quad_space = quad_space
        self._dx = dx(metadata=quad_params)

        if isinstance(target, UFLFunctionSpace):
            self._target_space = target._functionspace
        else:
            self._target_space = quad_space

        if source is None:
            quad_shape = get_numpy_shape(self._quad_space)
            source = Ndarrays(quad_shape)

        super().__init__(source, target)

    def __call__(self, array):
        Q = self._quad_space
        V = self._target_space
        q = function_set_local(Function(Q), array)
        if V != Q:
            q = project(q, V, dx=self._dx)
        return q


class UFLExprSpace_To_Numpy(PointMap):
    """This class is a point map that maps an expression to a NumPy array.

    The constructor creates a quadrature space for the source space, using the
    given quadrature parameters. The ``__call__`` method projects the input into the
    quadrature space and extracts the values as a NumPy array.

    Args:
        source (UFLExprSpace or UFLFunctionSpace): the UFL space to use as input.
        target (Ndarrays, optional): the target space to map to.
        quad_space (optional): the finite-element quadrature space to interpolate to.
        quad_params (dict, optional): parameters for the quadrature space.
        domain (optional): the UFL domain for the quadrature space.

    """

    def __init__(
        self, source, target=None, quad_space=None, quad_params=None, domain=None
    ):
        if quad_space is None:
            quad_space, quad_params = make_quadrature_spaces(
                source, quad_params=quad_params, domain=domain
            )
        self._quad_space = quad_space
        self._dx = dx(metadata=quad_params)

        if target is None:
            quad_shape = get_numpy_shape(self._quad_space)
            target = Ndarrays(quad_shape)

        super().__init__(source, target)

    def __call__(self, expr):
        Q = self._quad_space
        assert expr.ufl_shape == Q.ufl_element().value_shape()
        if isinstance(expr, Function) and Q == expr.function_space():
            func = expr
        else:
            func = project(expr, Q, dx=self._dx)
        array = function_get_local(func)
        return array


@register_covering(UFLExprSpace, Ndarrays)
@register_covering(UFLFunctionSpace, Ndarrays)
class Numpy_UFLFunctionSpace_Covering(Covering):

    # Base space should be a UFLExprSpace or a UFLFunctionSpace.
    def __init__(
        self,
        base_space,
        covering_space=None,
        domain=None,
        quad_params=None,
        **covering_params
    ):
        quad_space, quad_params = make_quadrature_spaces(
            base_space, quad_params=quad_params, domain=domain
        )
        self._quad_space = quad_space
        self._quad_params = quad_params

        if covering_space is None:
            quad_shape = get_numpy_shape(self._quad_space)
            covering_space = Ndarrays(quad_shape)

        super().__init__(base_space, covering_space, **covering_params)

    def covering_map(self):
        return Numpy_To_UFLFunctionSpace(
            self._covering_space,
            self._base_space,
            quad_space=self._quad_space,
            quad_params=self._quad_params,
        )

    def section_map(self):
        return UFLExprSpace_To_Numpy(
            self._base_space,
            self._covering_space,
            quad_space=self._quad_space,
            quad_params=self._quad_params,
        )


class To_UFLFunctionSpace(PointMap):
    """This class is a point map that projects a UFL expression to a function space.

    The ``__call__`` method projects the input onto the corresponding output
    function space.

    Args:
        source (UFLExprSpace or UFLFunctionSpace): the source space
        target (UFLFunctionSpace): the target space to project

    """

    def __init__(self, source, target):
        self._target_space = target._functionspace

        super().__init__(source, target)

    def __call__(self, expr):
        return project(expr, self._target_space)


@register_covering(UFLExprSpace, UFLFunctionSpace)
class UFLFunctionSpace_UFLExpr_Covering(Covering):
    # Covering space is a UFLFunctionSpace.
    def covering_map(self):
        return IdentityPointMap(self._covering_space, self._base_space)

    def section_map(self):
        return To_UFLFunctionSpace(self._base_space, self._covering_space)


@register_covering(UFLFunctionSpace, UFLFunctionSpace)
class UFLFunctionSpace_UFLFunctionSpace_Covering(Covering):
    def covering_map(self):
        return To_UFLFunctionSpace(self._covering_space, self._base_space)

    def section_map(self):
        return To_UFLFunctionSpace(self._base_space, self._covering_space)


def get_numpy_shape(function_space):
    return (
        Function(function_space)
        .vector()[:]
        .reshape(-1, *function_space.ufl_element().value_shape())
        .shape
    )



#!/usr/bin/python3
import numpy as np
from scipy.special import expit
import sys

ACTIVATIONS = {}
ACTIVATIONS["linear"] = lambda x: x
# ACTIVATIONS['softmax'] = lambda x: np.exp(x-np.max(x, axis=1)[:, None])/np.sum(np.exp(x-np.max(x, axis=1)[:, None]), axis=1)[:, None]
ACTIVATIONS["sigmoid"] = lambda x: expit(x)
ACTIVATIONS["tanh"] = lambda x: np.tanh(x)
ACTIVATIONS["relu"] = lambda x: x * (x > 0)
ACTIVATIONS["softplus"] = lambda x: np.log(np.exp(x) + 1)

ACT_DERIVATIVES = {}
ACT_DERIVATIVES["linear"] = lambda x, z: np.ones(z.shape)
# ACT_DERIVATIVES['softmax'] = lambda x, z: z*(1-z)
ACT_DERIVATIVES["sigmoid"] = lambda x, z: z * (1 - z)
ACT_DERIVATIVES["tanh"] = lambda x, z: 1 - z * z
ACT_DERIVATIVES["relu"] = lambda x, z: 1 * (x > 0)
ACT_DERIVATIVES["softplus"] = lambda x, z: 1 / (np.exp(-x) + 1)

ERRORS = {}
ERRORS["squared_error"] = lambda y_true, y_pred: np.mean((y_true - y_pred) ** 2)


def cross_entropy(y_true, y_pred):
    y_pred = y_pred.clip(0.00001, 1 - 0.00001)
    return -np.sum(y_true * np.log(y_pred) + (1 - y_true) * np.log(1 - y_pred))


def cross_entropy_der(y_true, y_pred):
    y_pred = y_pred.clip(0.00001, 1 - 0.00001)
    return -y_true / y_pred + (1 - y_true) / (1 - y_pred)


ERRORS["cross_entropy"] = cross_entropy

ERR_DERIVATIVES = {}
ERR_DERIVATIVES["squared_error"] = lambda y_true, y_pred: -(y_true - y_pred) / len(
    y_true.flatten()
)
ERR_DERIVATIVES["cross_entropy"] = cross_entropy_der


class Layer(object):
    def __init__(self, num_outputs, input_shape, activation="sigmoid", weights=None):
        self.input_shape = input_shape
        self.num_outputs = num_outputs
        self.activation = activation

        self.output = np.zeros((self.num_outputs))
        if len(input_shape) > 1:
            print(
                "Layer() error: input_shape must be one-dimensional for this layer type"
            )
            sys.exit(1)

        if weights is None:
            # Generate weights between -0.3 and 0.3
            weights = np.random.random_sample(
                (self.input_shape[0] + 1, self.num_outputs)
            )
            weights = weights * 0.6 - 0.3
        self.weights = weights
        self.deltaw = np.zeros_like(self.weights)

    def create_copy(self, weights=None):
        if weights is None:
            weights = self.weights.copy()
        elif isinstance(weights, str) and weights == "random":
            weights = None
        else:
            weights = weights.copy()
            assert weights.shape == self.weights.shape

        return Layer(self.num_outputs, self.input_shape, self.activation, weights)

    def forward_pass(self, data):
        self.input = np.append(data, np.ones(data.shape[0])[:, None], axis=1)
        self.act = np.dot(self.input, self.weights)
        self.output = ACTIVATIONS[self.activation](self.act)

    def compute_jacobian(self):
        # I removed the bottom row from the weights, because they correspond to the bias terms.
        act_der = ACT_DERIVATIVES[self.activation](self.act, self.output)
        my_jacobian = act_der[:, :, None] * np.transpose(self.weights[:-1, :])
        return my_jacobian

    def backward_pass(self, next_layer, err_der=None):
        if err_der is None:
            err_der = np.dot(next_layer.err, np.transpose(next_layer.weights))[:, :-1]
        act_der = ACT_DERIVATIVES[self.activation](self.act, self.output)
        self.err = err_der * act_der
        self.weight_gradient = np.dot(self.input.transpose(), self.err)

    def update_weights(self, learning_rate, momentum):
        self.deltaw = momentum * self.deltaw - learning_rate * self.weight_gradient
        self.weights += self.deltaw


class Network(object):
    def __init__(self, layers=None, filename=None):
        self.error = "squared_error"
        if filename is None:
            self.layers = layers
        else:
            self.load(filename)
        if len(self.layers) == 0:
            print("Network() error: len(layers) must be > 0")
            sys.exit(1)
        self.num_inputs = self.layers[0].input_shape[0]
        self.num_outputs = self.layers[-1].num_outputs

    @staticmethod
    def create_layers(
        num_inputs,
        num_outputs,
        output_activation,
        layer_sizes,
        layer_activation,
        seed=None,
    ):
        layers = []
        prev_output_shape = (num_inputs,)
        np.random.seed(seed)
        for n in layer_sizes:
            l = Layer(
                num_outputs=n,
                input_shape=prev_output_shape,
                activation=layer_activation,
            )
            layers.append(l)
            prev_output_shape = (l.num_outputs,)
        layers.append(
            Layer(
                num_outputs=num_outputs,
                input_shape=prev_output_shape,
                activation=output_activation,
            )
        )
        return layers

    def create_copy(self, weights=None):
        layers = []
        if weights == "random":
            for l in self.layers:
                l = l.create_copy(weights=weights)
                layers.append(l)
            return Network(layers)

        if weights is None:
            weights = self.get_weights(copy=False)

        assert len(weights) == len(self.layers)
        for l, w in zip(self.layers, weights):
            l = l.create_copy(weights=w)
            layers.append(l)
        return Network(layers)

    def get_weights(self, copy=True):
        weights = []
        for l in self.layers:
            weights.append(l.weights if not copy else l.weights.copy())
        return weights

    def set_weights(self, weights):
        assert len(weights) == len(self.layers)
        for l, w in zip(self.layers, weights):
            l.weights = w

    def train(self, x, y, learning_rate=0.001, momentum=0, debug=False):
        # Forward pass
        self.predict(x, debug=debug)

        # Backward Pass
        self.derivative_calculations(
            ERR_DERIVATIVES[self.error](y, self.layers[-1].output), debug=debug
        )

        # Update weights
        self.update_weights(learning_rate=learning_rate, momentum=momentum, debug=debug)

        return self.layers[-1].output, ERRORS[self.error](y, self.layers[-1].output)

    def derivative_calculations(self, err_der, debug=False):
        if debug:
            print("Backward Pass (err)")

        l = self.layers[-1]
        l.backward_pass(None, err_der)
        if debug:
            print("Layer %d: %s" % (len(self.layers) - 1, l.err.shape))
        for i in range(len(self.layers) - 2, -1, -1):
            l = self.layers[i]
            l.backward_pass(self.layers[i + 1])
            if debug:
                print("Layer %d: %s" % (i, l.err.shape))

        l = self.layers[0]
        self.input_err = np.dot(l.err, np.transpose(l.weights))[:, :-1]
        self.param_err = []
        for l in self.layers:
            self.param_err.append(l.weight_gradient)

        if debug:
            print("Input: %s" % (self.input_err.shape,))
            print()

    def update_weights(self, learning_rate=0.001, momentum=0, debug=False):
        if debug:
            print("Updating weights (deltaw)")
        for l in self.layers:
            l.update_weights(learning_rate=learning_rate, momentum=momentum)
        if debug:
            print("Successfully updated weights. Quitting now...")
            sys.exit(1)

    def predict(self, x, debug=False):
        self.layers[0].forward_pass(x)
        if debug:
            print("Forward Pass (input -> output)")
            print(
                "Layer %d: %s -> %s"
                % (0, self.layers[0].input.shape, self.layers[0].output.shape)
            )
        for i in range(1, len(self.layers)):
            l = self.layers[i]
            l.forward_pass(self.layers[i - 1].output)
            if debug:
                print("Layer %d: %s -> %s" % (i, l.input.shape, l.output.shape))
        if debug:
            print()
        return self.layers[-1].output

    def jacobian_calculations(self):
        jac = self.layers[0].compute_jacobian()
        for i in range(1, len(self.layers)):
            l = self.layers[i]
            ljac = l.compute_jacobian()
            jac = np.matmul(ljac, jac)
        self.jacobian = jac
        self.p_jacobian = None

    def summary(self):
        s = "============================\n"
        s += "    Network Summary\n\n"
        s += "(input_size) -> output_size\n\n"
        for i, l in enumerate(self.layers):
            s += "Layer %d: %s -> %s\n" % (i, l.input_shape, l.num_outputs)
        s += "===========================\n"
        return s

    def save(self, filename):
        import pickle

        with open(filename, "wb") as outputFile:
            pickle.dump(self.layers, outputFile, pickle.HIGHEST_PROTOCOL)

    def load(self, filename):
        import pickle

        with open(filename, "rb") as inputFile:
            self.layers = pickle.load(inputFile)


class PlapNetwork(object):
    def __init__(self, p=2, dims=2, input_just_vector=False, output_vector=False):
        self.p = p
        self.dims = dims
        self.deltap = 0
        self.output_vector = output_vector
        self.input_just_vector = input_just_vector
        self.num_inputs = self.dims if self.input_just_vector else 1 + self.dims
        self.num_outputs = self.dims if self.output_vector else 1

    @staticmethod
    def create_layers(
        num_inputs, num_outputs, output_activation, layer_sizes, layer_activation, seed
    ):
        raise NotImplementedError("PlapNetwork doesn't have this functionality")

    def train(self, x, y, learning_rate=0.001, momentum=0, debug=False):
        # Forward pass
        self.predict(x, debug=debug)

        # Backward Pass
        self.derivative_calculations(
            ERR_DERIVATIVES[self.error](y, l.output), debug=debug
        )

        # Update weights
        self.update_weights(debug=debug, learning_rate=learning_rate, momentum=momentum)

        return self.output, ERRORS[self.error](y, self.output)

    def create_copy(self, p=None, plap_network=None):
        layers = []

        if plap_network is not None:
            if p is not None:
                raise ValueError("It's invalid to define both p and plap_network")

            return PlapNetwork(
                obj.p,
                dims=plap_network.dims,
                input_just_vector=plap_network.input_just_vector,
                output_vector=plap_network.output_vector,
            )

        if p == "random":
            return PlapNetwork(
                2 + np.random.rand(),
                dims=self.dims,
                input_just_vector=self.input_just_vector,
                output_vector=self.output_vector,
            )

        if p is None:
            p = self.p
        return PlapNetwork(
            p=p,
            dims=self.dims,
            input_just_vector=self.input_just_vector,
            output_vector=self.output_vector,
        )

    def derivative_calculations(self, err_der, debug=False):
        self.jacobian_calculations()
        self.err_der = err_der
        if self.output_vector:
            self.input_err = np.matmul(self.err_der[:, None, :], self.jacobian)
        else:
            self.input_err = self.err_der * self.jacobian[:, 0, :]

        axes = np.arange(len(self.err_der.shape))
        self.param_err = np.tensordot(self.err_der, self.p_jacobian, (axes, axes))[0]
        if self.param_err.size == 1:
            self.param_err = self.param_err.flat[0]

        if not self.input_just_vector:
            # Add a column of zeros to input_err to take the place of the function value, which has a gradient of 0.
            zeros = np.zeros((self.input_err.shape[0], self.input_err.shape[1] + 1))
            zeros[:, 1:] = self.input_err
            self.input_err = zeros

    def update_weights(self, learning_rate=0.001, momentum=0, debug=False):
        self.deltap = momentum * self.deltap - learning_rate * np.dot(
            self.err_der.transpose(), self.p_jacobian
        )
        self.p += self.deltap

    def predict(self, x, debug=False):

        if self.input_just_vector:
            self.input = x[:, :]
        else:
            # The first column of x is the function value, and the other columns are the gradient.
            self.input = x[:, 1:]

        self.activation = np.sum(self.input * self.input, axis=1) + 1e-12
        self.activation = self.activation[:, None]

        self.output = self.activation ** ((self.p - 2) / 2)

        if self.output_vector:
            # Multiply the output magnitudes by gradu.
            self.output = self.output * self.input
        return self.output

    def jacobian_calculations(self):
        a = self.activation
        g = self.input
        q = self.p - 2

        if self.output_vector:

            self.jacobian = np.matmul(
                (q * (a ** (q / 2 - 1)) * g)[:, :, None], g[:, None, :]
            )

            # Then add to diagonal
            diag = np.arange(self.dims)
            self.jacobian[:, diag, diag] += a ** (q / 2)

            if not self.input_just_vector:
                # Add a column of zeros at the beginning to represent the function value, which has a gradient of 0.
                jac = np.zeros((self.jacobian.shape[0], self.jacobian.shape[1] + 1))
                jac[:, 1:] = self.jacobian
                self.jacobian = jac

        else:
            self.jacobian = a ** (q / 2 - 1) * q * g
            self.jacobian = self.jacobian[:, None, :]

        # self.p_jacobian = self.output*np.log(self.activation)[:, None]
        self.p_jacobian = self.output * 0.5 * np.log(a)
        self.p_jacobian = np.expand_dims(self.p_jacobian, axis=-1)

    def summary(self):
        s = "============================\n"
        s += "    PlapNetwork Summary\n\n"
        s += "        p = %g\n" % self.p
        s += "     dims = %d\n" % self.dims
        s += "===========================\n"
        return s

    def save(self, filename):
        raise NotImplementedError("PlapNetwork doesn't have this functionality")

    def load(self, filename):
        raise NotImplementedError("PlapNetwork doesn't have this functionality")
